'use strict';
/* This STable mode flag is set if the type is parametric (and so can be
 * parameterized). */
exports.PARAMETRIC_TYPE = 16;

/* This STable mode flag is set if the type is a parameterization of some
 * parametric type. */
exports.PARAMETERIZED_TYPE = 32;

exports.METHOD_CACHE_AUTHORITATIVE = 4;
